package de.fraunhofer.iem.dataprovider.taskManager.tasks

import de.fraunhofer.iem.dataprovider.gitlab.RepositoryRepository
import de.fraunhofer.iem.dataprovider.sarif.Sarif
import de.fraunhofer.iem.dataprovider.sarif.asDbObject
import de.fraunhofer.iem.dataprovider.sarif.getSarifFromFilePath
import de.fraunhofer.iem.dataprovider.toolResult.ToolRunRepository
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import org.springframework.transaction.annotation.Transactional
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import java.util.*

sealed class ProcessTask : Task() {
    protected abstract val flags: Array<String>
    protected abstract val execPath: String


    private val mainScope = CoroutineScope(Dispatchers.Default)
    override suspend fun execute() {
        if (isExecutable(execPath)) {
            val process = ProcessBuilder(execPath, *flags).start()

            process.onExit().thenApply { p1 ->
                mainScope.launch {
                    handleProcessReturn(p1)
                }
            }
        } else {
            logger.warn("Given execPath is not an executable $execPath.")
            // TODO: we should probably throw an exception here.
        }
    }

    private fun isExecutable(filePath: String): Boolean {
        val path = Paths.get(filePath)
        return Files.isExecutable(path)
    }

    abstract suspend fun handleProcessReturn(p: Process)

}
@Transactional
sealed class SarifTask : ProcessTask() {
    protected abstract val repoId: UUID
    protected abstract val repository: RepositoryRepository
    protected abstract val toolRunRepository: ToolRunRepository
    protected abstract val resultPath: Path


    override suspend fun handleProcessReturn(p: Process) {

        // TODO: check process exit codes. Current problem, if an analysis
        //  has findings it might return an exit code != 0 even tho the process finished correctly

        logger.info("Handle Process return in $javaClass")
        val sarifResult = getSarifFromFilePath(resultPath)
//        val repo = repository.findById(repoId)
        val sarifDb = sarifResult.asDbObject()
        val results = toolRunRepository.saveAll(sarifDb)
        results.forEach {
            val dbres = toolRunRepository.findById(it.id!!).get()
//            Hibernate.initialize(dbres.toolResults)
            logger.info(dbres.toString())
        }
//        repo.ifPresent {
//            Hibernate.initialize(it.toolRuns) // TODO: this fails because the session seems to be closed and add won't work because toolRuns are loaded lazily
//            it.toolRuns.addAll(results)
//            repository.save(it)
//        }
        sendResult(sarifResult)
    }

    abstract suspend fun sendResult(sarif: Sarif)

}
